import { BehaviorSubject, Subject, Observable } from 'rxjs';
import { share } from 'rxjs/operators';

import { Injectable, OnDestroy } from '@angular/core';

import i18next, {
  InitOptions,
  Module,
  Newable,
  ThirdPartyModule,
  TOptions,
  StringMap
} from 'i18next';

import { LogService } from '@ucap/ng-logger';

import { I18nextEvents } from '../types/i18next.event';

@Injectable()
export class I18nService implements OnDestroy {
  readonly initialized$: Observable<InitOptions>;
  readonly loaded$: Observable<boolean>;
  readonly failedLoading$: Observable<{
    lng: string;
    ns: string;
    msg: string;
  }>;
  readonly missingKey$: Observable<{
    lngs: string[];
    ns: string;
    key: string;
    res: string;
  }>;
  readonly added$: Observable<{ lng: string; ns: string }>;
  readonly removed$: Observable<{ lng: string; ns: string }>;
  readonly languageChanged$: Observable<string>;

  get options(): InitOptions {
    return i18next.options;
  }

  get currentLng(): string {
    return this._currentLng;
  }

  private readonly initializedSubject = new BehaviorSubject<InitOptions>(null);
  private readonly loadedSubject = new BehaviorSubject<boolean>(false);
  private readonly failedLoadingSubject = new Subject<{
    lng: string;
    ns: string;
    msg: string;
  }>();
  private readonly missingKeySubject = new Subject<{
    lngs: string[];
    ns: string;
    key: string;
    res: string;
  }>();
  private readonly addedSubject = new Subject<{ lng: string; ns: string }>();
  private readonly removedSubject = new Subject<{ lng: string; ns: string }>();
  private readonly languageChangedSubject = new BehaviorSubject<string>(null);

  // tslint:disable-next-line: variable-name
  private _currentLng: string;

  constructor(private logService: LogService) {
    this.initialized$ = this.initializedSubject.asObservable().pipe(share());
    this.loaded$ = this.loadedSubject.asObservable().pipe(share());
    this.failedLoading$ = this.failedLoadingSubject
      .asObservable()
      .pipe(share());
    this.missingKey$ = this.missingKeySubject.asObservable().pipe(share());
    this.added$ = this.addedSubject.asObservable().pipe(share());
    this.removed$ = this.removedSubject.asObservable().pipe(share());
    this.languageChanged$ = this.languageChangedSubject
      .asObservable()
      .pipe(share());
  }

  ngOnDestroy(): void {
    this.initializedSubject.next(null);
    this.initializedSubject.complete();

    this.loadedSubject.next(null);
    this.loadedSubject.complete();

    this.failedLoadingSubject.next(null);
    this.failedLoadingSubject.complete();

    this.missingKeySubject.next(null);
    this.missingKeySubject.complete();

    this.addedSubject.next(null);
    this.addedSubject.complete();

    this.removedSubject.next(null);
    this.removedSubject.complete();

    this.languageChangedSubject.next(null);
    this.languageChangedSubject.complete();
  }

  use<T extends Module>(
    module: T | Newable<T> | ThirdPartyModule[] | Newable<ThirdPartyModule>[]
  ) {
    i18next.use(module);
    return this;
  }

  init(options?: InitOptions) {
    options = options || {};

    this.subscribeEvents();

    return i18next.init(options);
  }

  changeLanguage(lng: string) {
    const self = this;
    return new Promise<void>((resolve, reject) => {
      if (lng === self.currentLng) {
        resolve();
        return;
      }

      i18next
        .changeLanguage(lng)
        .then(v => {
          self.logService.info('changeLanguage', lng);
          resolve();
        })
        .catch(reason => {
          reject(reason);
        });
    });
  }

  loadNamespaces(ns: string | string[]) {
    return i18next.loadNamespaces(ns);
  }

  loadLanguages(lngs: string | string[]) {
    return i18next.loadLanguages(lngs);
  }

  reloadResources(lngs?: string | string[], ns?: string | string[]) {
    return i18next.reloadResources(lngs, ns);
  }

  getResource(
    lng: string,
    ns: string,
    key: string,
    options?: { keySeparator?: string }
  ) {
    return i18next.getResource(lng, ns, key, options);
  }

  addResource(
    lng: string,
    ns: string,
    key: string,
    value: string,
    options?: { keySeparator?: string }
  ) {
    i18next.addResource(lng, ns, key, value, options);
  }

  addResources(lng: string, ns: string, resources: any) {
    i18next.addResources(lng, ns, resources);
  }

  addResourceBundle(
    lng: string,
    ns: string,
    resources: any,
    deep?: boolean,
    overwrite?: boolean
  ) {
    i18next.addResourceBundle(lng, ns, resources, deep, overwrite);
  }

  hasResourceBundle(lng: string, ns: string) {
    return i18next.hasResourceBundle(lng, ns);
  }

  getResourceBundle(lng: string, ns: string) {
    return i18next.getResourceBundle(lng, ns);
  }

  removeResourceBundle(lng: string, ns: string) {
    i18next.removeResourceBundle(lng, ns);
  }

  t(
    key: string | string[],
    optionsOrDefault?: string | TOptions<StringMap>
  ): string;
  t(
    key: string | string[],
    optionsOrDefaultValue?: string,
    options?: string | TOptions<StringMap>
  ): string {
    return i18next.t(key, optionsOrDefaultValue, options);
  }

  format(value: any, format?: string, lng?: string) {
    return i18next.format(value, format, lng);
  }

  exists(key: string | string[], options: TOptions<StringMap>) {
    return i18next.exists(key, options);
  }

  getFixedT(lng: string | string[], ns?: string | string[]) {
    return i18next.getFixedT(lng, ns);
  }

  setDefaultNamespace(ns: string) {
    i18next.setDefaultNamespace(ns);
  }

  dir(lng?: string) {
    return i18next.dir(lng);
  }

  private subscribeEvents() {
    // tslint:disable-next-line: variable-name
    const __this = this;

    i18next.on(I18nextEvents.initialized, options => {
      __this.initializedSubject.next(options);
    });
    i18next.on(I18nextEvents.loaded, loaded => {
      __this.loadedSubject.next(loaded);
    });
    i18next.on(I18nextEvents.failedLoading, (lng, ns, msg) => {
      __this.failedLoadingSubject.next({ lng, ns, msg });
    });
    i18next.on(I18nextEvents.languageChanged, lng => {
      __this.languageChangedSubject.next(lng);
      this._currentLng = lng;
    });
    i18next.on(I18nextEvents.missingKey, (lngs, ns, key, res) => {
      __this.missingKeySubject.next({ lngs, ns, key, res });
    });
    i18next.on(I18nextEvents.added, (lng, ns) => {
      __this.addedSubject.next({ lng, ns });
    });
    i18next.on(I18nextEvents.removed, (lng, ns) => {
      __this.removedSubject.next({ lng, ns });
    });
  }
}
